/* Mul 32 x 32
 * ------------------------
 * This is a little demo how to multiply two unsigned 32 bit values using only 16bit multiplications,
 * casts to uint32_t respectively uint64_t and shift operations.
 *
 * Note that this is just a quick-and-dirty example to show how to use it.
 *
 * Have a look at the notes describing the algorithm:
 *
 * Sunshine, November 2k17
 * www.sunshine2k.de | www.bastian-molkenthin.de
 */

#include <stdio.h>
#include <stdint.h>

typedef struct TestCaseStructTag
{
	uint32_t valueA;
	uint32_t valueB;
} TestCaseStruct;

TestCaseStruct testCases[] = {
	{ 0x00000002, 0xFFFFFFFF },
	{ 0x0000FFFF, 0xFFFFFFFF },
	{ 0x10101010, 0xFFFFFFFF },
	{ 0x10101010, 0x10101010 },
	{ 0xFFFFFFFF, 0x0002000A },
	{ 0xFFFFFFFF, 0x0002000A },
	{ 0xFFFFFFFF, 0xFFFFFFFF },
	{ 0x1FFF1FFF, 0xFFFEFFFE },
	{ 0x00000000, 0xFFFFFFFF },
	{ 0xFFFFFFFE, 0xFFFFFFFE },
	{ 0x00003300, 0x000000A0 },
	{ 0x00003300, 0x001B0000 },
	{ 0x01020000, 0x001B0000 },
	{ 0x01020000, 0x0000EE30 },
};

/*
va * vb = (Ha << 16 + La) * (Hb << 16 + Lb)
        = (Ha * Hb << 32) * ((Ha << 16) * Lb) + ((Hb << 16) * La) + La*Lb
*/
uint64_t mymul_1(uint32_t a, uint32_t b)
{
	uint16_t va_high = (a >> 16) & 0xFFFF;
	uint16_t va_low = a & 0xFFFF;
		
	uint16_t vb_high = (b >> 16) & 0xFFFF;
	uint16_t vb_low = b & 0xFFFF;

	uint64_t mul_high_high = (uint64_t)((uint32_t)(va_high * vb_high))<< 32;
	uint64_t mul_high_low  = (uint64_t)((uint32_t)(va_high << 16)) * vb_low;
	uint64_t mul_low_high  = (uint64_t)((uint32_t)(vb_high << 16)) * va_low;
	uint32_t mul_low_low   = (uint32_t)(va_low * vb_low);

	uint64_t res = 0;
	
	res  = mul_high_high;
	res += mul_high_low;
	res += mul_low_high;
	res += mul_low_low;

	return res;
}

uint64_t mymul_2(uint32_t a, uint32_t b)
{
	uint16_t va_high = (a >> 16) & 0xFFFF;
	uint16_t va_low = a & 0xFFFF;
		
	uint16_t vb_high = (b >> 16) & 0xFFFF;
	uint16_t vb_low = b & 0xFFFF;

	uint32_t mul_high_high = (uint32_t)(va_high * vb_high);
	uint32_t mul_high_low  = (uint32_t)(va_high * vb_low);
	uint32_t mul_low_high  = (uint32_t)(vb_high * va_low);
	uint32_t mul_low_low   = (uint32_t)(va_low * vb_low);

	uint64_t res = 0;
	res  = ((uint64_t)mul_high_high) << 32;
	res += ((uint64_t)mul_high_low + mul_low_high) << 16;
	res += mul_low_low;

	return res;
}

uint16_t RunTestcases()
{
	uint64_t expResultVal, mymul1ResultVal, mymul2ResultVal;
	uint16_t testCaseLoop;
	uint16_t numOfFailedTestcases;

	numOfFailedTestcases = 0;
	for (testCaseLoop = 0; testCaseLoop < sizeof(testCases) / sizeof(TestCaseStruct); testCaseLoop++)
	{
		expResultVal = (int64_t)(testCases[testCaseLoop].valueA) * (int64_t)(testCases[testCaseLoop].valueB);
		mymul1ResultVal = mymul_1(testCases[testCaseLoop].valueA, testCases[testCaseLoop].valueB);
		mymul2ResultVal = mymul_2(testCases[testCaseLoop].valueA, testCases[testCaseLoop].valueB);

		if ((expResultVal == mymul1ResultVal) && (expResultVal == mymul2ResultVal))
		{
			printf("x");
		}
		else
		{
			if (expResultVal != mymul1ResultVal)
			{
				printf("\nFail - 1! ValueA = %.8lX, ValueB = %.8lX, Expected result = %.16llX, Actual Result: %.16llX\n", 
					testCases[testCaseLoop].valueA, testCases[testCaseLoop].valueB, expResultVal, mymul1ResultVal);
			}
			else
			{
				printf("\nFail - 2! ValueA = %.8lX, ValueB = %.8lX, Expected result = %.16llX, Actual Result: %.16llX\n", 
					testCases[testCaseLoop].valueA, testCases[testCaseLoop].valueB, expResultVal, mymul2ResultVal);
			}
			numOfFailedTestcases++;
		}
	}
	printf("\n");
	return numOfFailedTestcases;
}

int main(int argc, char* argv[])
{
	printf("Unsigned 32x32bit multiplication test application\n");
	printf("www.sunshine2k.de || www.bastian-molkenthin.de\n");
	printf("-------------------------------------------------\n");

	RunTestcases();
	
	printf("\nPrint any key to quit...\n");
	getchar();
}
   

